"""
Unit tests for the curve server mechanism class.
"""

import pytest

from azmq.crypto import (
    crypto_box,
    crypto_box_afternm,
    crypto_box_beforenm,
    crypto_secretbox,
    curve_gen_keypair,
)
from azmq.mechanisms.curve_server import CurveServer
from azmq.errors import ProtocolError

from ...conftest import requires_libsodium


@pytest.fixture
def mechanism():
    public_key, secret_key = curve_gen_keypair()
    mechanism = CurveServer(
        public_key=public_key,
        secret_key=secret_key,
    )()
    mechanism.rc, mechanism.rs = curve_gen_keypair()
    mechanism.rcp, mechanism.rsp = curve_gen_keypair()
    mechanism.kp = crypto_box_beforenm(
        pk=mechanism.rcp,
        sk=mechanism.sp,
    )
    mechanism.rkp = crypto_box_beforenm(
        pk=mechanism.cp,
        sk=mechanism.rsp,
    )
    return mechanism


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_hello_invalid(reader, mechanism):
    reader.write(b'\x04\x06\x05HELLO')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_hello(reader=reader)


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_hello_invalid_version(reader, mechanism):
    reader.write(b'\x04\xc8\x05HELLO' + b'\0' * 194)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_hello(reader=reader)


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_hello_invalid_box(reader, mechanism):
    reader.write(b'\x04\xc8\x05HELLO\1\0' + b'\0' * 192)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_hello(reader=reader)


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_invalid(reader, mechanism):
    reader.write(b'\x04\x09\x08INITIATE')
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=b'\0' * 32,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_invalid_cookie(reader, mechanism):
    reader.write(b'\x04\x71\x08INITIATE' + b'\0' * 104)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=b'\0' * 32,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_mismatching_cookie(reader, mechanism):
    cookie_nonce = b'\0' * 16
    cookie_key = b'\0' * 32
    cookie = crypto_secretbox(
        b'\0' * 32 + mechanism.sp,
        nonce=b'COOKIE--' + cookie_nonce,
        k=cookie_key,
    )
    nonce = b'\0' * 8
    reader.write(b'\x04\x71\x08INITIATE' + cookie_nonce + cookie + nonce)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=cookie_key,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_invalid_vouch_box(reader, mechanism):
    cookie_nonce = b'\0' * 16
    cookie_key = b'\0' * 32
    cookie = crypto_secretbox(
        mechanism.rcp + mechanism.sp,
        nonce=b'COOKIE--' + cookie_nonce,
        k=cookie_key,
    )
    nonce = b'\0' * 8
    reader.write(b'\x04\x71\x08INITIATE' + cookie_nonce + cookie + nonce)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=cookie_key,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_invalid_vouch_size(reader, mechanism):
    cookie_nonce = b'\0' * 16
    cookie_key = b'\0' * 32
    cookie = crypto_secretbox(
        mechanism.rcp + mechanism.sp,
        nonce=b'COOKIE--' + cookie_nonce,
        k=cookie_key,
    )
    nonce = b'\0' * 8
    plain = b'\0' * 47
    box = crypto_box_afternm(
        plain,
        nonce=b'CurveZMQINITIATE' + nonce,
        k=mechanism.rkp,
    )
    reader.write(b'\x04\xb1\x08INITIATE' + cookie_nonce + cookie + nonce + box)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=cookie_key,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_invalid_vouch(reader, mechanism):
    cookie_nonce = b'\0' * 16
    cookie_key = b'\0' * 32
    cookie = crypto_secretbox(
        mechanism.rcp + mechanism.sp,
        nonce=b'COOKIE--' + cookie_nonce,
        k=cookie_key,
    )
    nonce = b'\0' * 8
    plain = b'\0' * 48
    box = crypto_box_afternm(
        plain,
        nonce=b'CurveZMQINITIATE' + nonce,
        k=mechanism.rkp,
    )
    reader.write(b'\x04\xb1\x08INITIATE' + cookie_nonce + cookie + nonce + box)
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=cookie_key,
        )


@requires_libsodium
@pytest.mark.asyncio
async def test_read_curve_initiate_mismatching_vouch(reader, mechanism):
    cookie_nonce = b'\0' * 16
    cookie_key = b'\0' * 32
    cookie = crypto_secretbox(
        mechanism.rcp + mechanism.sp,
        nonce=b'COOKIE--' + cookie_nonce,
        k=cookie_key,
    )
    vouch_nonce = b'\0' * 16
    vouch_box = crypto_box(
        b'0' * 64,
        # mechanism.rcp + mechanism.c,
        nonce=b'VOUCH---' + vouch_nonce,
        pk=mechanism.cp,
        sk=mechanism.rs,
    )
    plain = mechanism.rc + vouch_nonce + vouch_box
    nonce = b'\0' * 8
    box = crypto_box_afternm(
        plain,
        nonce=b'CurveZMQINITIATE' + nonce,
        k=mechanism.rkp,
    )
    reader.write(
        b'\x06\x00\x00\x00\x00\x00\x00\xff\x02\x08INITIATE' + cookie_nonce +
        cookie + nonce + box,
    )
    reader.seek(0)

    with pytest.raises(ProtocolError):
        await mechanism._read_curve_initiate(
            reader=reader,
            cookie_key=cookie_key,
        )



"""
Seucrity integration tests.
"""

import asyncio
import pytest

import azmq

from azmq.mechanisms import (
    CurveClient,
    CurveServer,
    PlainClient,
    PlainServer,
)
from azmq.crypto import curve_gen_keypair
from azmq.errors import (
    ZAPAuthenticationFailure,
    ZAPError,
    ZAPInternalError,
    ZAPTemporaryError,
)
from azmq.zap import (
    BaseZAPAuthenticator,
    ZAPAuthenticator,
    ZAPClient,
    ZAP_INPROC_ENDPOINT,
)

from ..conftest import requires_libsodium


ENDPOINT = 'tcp://127.0.0.1:5000'


def onesec(awaitable):
    """
    Causes a normally blocking call to timeout after a while.

    :param awaitable: The awaitable to wrap.
    :returns: A decorated awaitable that times out.
    """
    return asyncio.wait_for(awaitable, 1)


@pytest.mark.asyncio
async def test_ip_allow(event_loop):
    async with azmq.Context() as context:
        authenticator = ZAPAuthenticator(context)
        authenticator.allow('127.0.0.1')
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])
            message = await onesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']
            await rep_socket.send_multipart([b'my', b'response'])
            message = await onesec(req_socket.recv_multipart())
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@pytest.mark.asyncio
async def test_ip_deny(event_loop):
    async with azmq.Context() as context:
        authenticator = ZAPAuthenticator(context)
        authenticator.deny('127.0.0.1')
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])

            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(rep_socket.recv_multipart(), 0.25)

        finally:
            req_socket.close()
            rep_socket.close()


@pytest.mark.asyncio
async def test_plain_valid_password(event_loop):
    async with azmq.Context() as context:
        authenticator = ZAPAuthenticator(context)
        authenticator.add_user(username='user', password='pwd')
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=PlainClient(username='user', password='pwd'),
        )
        rep_socket = context.socket(azmq.REP, mechanism=PlainServer())

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])
            message = await onesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']
            await rep_socket.send_multipart([b'my', b'response'])
            message = await onesec(req_socket.recv_multipart())
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@pytest.mark.asyncio
async def test_plain_unknown_username(event_loop):
    async with azmq.Context() as context:
        authenticator = ZAPAuthenticator(context)
        authenticator.add_user(username='user', password='pwd')
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=PlainClient(username='user2', password='pwd'),
        )
        rep_socket = context.socket(azmq.REP, mechanism=PlainServer())

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])

            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(rep_socket.recv_multipart(), 0.25)

        finally:
            req_socket.close()
            rep_socket.close()


@pytest.mark.asyncio
async def test_plain_invalid_password(event_loop):
    async with azmq.Context() as context:
        authenticator = ZAPAuthenticator(context)
        authenticator.add_user(username='user', password='pwd')
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=PlainClient(username='user', password='pwd2'),
        )
        rep_socket = context.socket(azmq.REP, mechanism=PlainServer())

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])

            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(rep_socket.recv_multipart(), 0.25)

        finally:
            req_socket.close()
            rep_socket.close()


@requires_libsodium
@pytest.mark.asyncio
async def test_curve_valid_key(event_loop):
    async with azmq.Context() as context:
        c_public_key, c_secret_key = curve_gen_keypair()
        s_public_key, s_secret_key = curve_gen_keypair()
        authenticator = ZAPAuthenticator(context)
        authenticator.add_authorized_key(key=c_public_key)
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=CurveClient(
                public_key=c_public_key,
                secret_key=c_secret_key,
                server_key=s_public_key,
            ),
        )
        rep_socket = context.socket(
            azmq.REP,
            mechanism=CurveServer(
                public_key=s_public_key,
                secret_key=s_secret_key,
            ),
        )

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])
            message = await onesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']
            await rep_socket.send_multipart([b'my', b'response'])
            message = await onesec(req_socket.recv_multipart())
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@requires_libsodium
@pytest.mark.asyncio
async def test_curve_invalid_key(event_loop):
    async with azmq.Context() as context:
        c_public_key, c_secret_key = curve_gen_keypair()
        s_public_key, s_secret_key = curve_gen_keypair()
        authenticator = ZAPAuthenticator(context)
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=CurveClient(
                public_key=c_public_key,
                secret_key=c_secret_key,
                server_key=s_public_key,
            ),
        )
        rep_socket = context.socket(
            azmq.REP,
            mechanism=CurveServer(
                public_key=s_public_key,
                secret_key=s_secret_key,
            ),
        )

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])

            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(rep_socket.recv_multipart(), 0.25)

        finally:
            req_socket.close()
            rep_socket.close()


@requires_libsodium
@pytest.mark.asyncio
async def test_incompatible_mechanisms(event_loop):
    async with azmq.Context() as context:
        c_public_key, c_secret_key = curve_gen_keypair()
        s_public_key, s_secret_key = curve_gen_keypair()
        authenticator = ZAPAuthenticator(context)
        authenticator.add_authorized_key(key=c_public_key)
        context.set_zap_authenticator(authenticator)
        req_socket = context.socket(
            azmq.REQ,
            mechanism=CurveClient(
                public_key=c_public_key,
                secret_key=c_secret_key,
                server_key=s_public_key,
            ),
        )
        rep_socket = context.socket(azmq.REP)

        try:
            req_socket.connect(ENDPOINT)
            rep_socket.bind(ENDPOINT)

            await req_socket.send_multipart([b'my', b'request'])

            with pytest.raises(asyncio.TimeoutError):
                await asyncio.wait_for(rep_socket.recv_multipart(), 0.25)

        finally:
            req_socket.close()
            rep_socket.close()


@pytest.mark.asyncio
async def test_zap_successful_authentication(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            return 'bob', {b'foo': b'bar'}

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                username, metadata = await asyncio.wait_for(
                    zap_client.authenticate(
                        domain='domain',
                        address='127.0.0.1',
                        identity=b'bob',
                        mechanism=b'CURVE',
                        credentials=[b'mycred', b'value'],
                    ),
                    1,
                )

    assert username == 'bob'
    assert metadata == {b'foo': b'bar'}


@pytest.mark.asyncio
async def test_zap_temporary_error(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            raise ZAPTemporaryError("Some error")

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                with pytest.raises(ZAPTemporaryError) as error:
                    await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert error.value.text == 'Some error'
    assert error.value.code == 300


@pytest.mark.asyncio
async def test_zap_authentication_failure(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            raise ZAPAuthenticationFailure("Some error")

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                with pytest.raises(ZAPAuthenticationFailure) as error:
                    await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert error.value.code == 400


@pytest.mark.asyncio
async def test_zap_internal_error(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            raise RuntimeError

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                with pytest.raises(ZAPInternalError) as error:
                    await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert error.value.code == 500


@pytest.mark.asyncio
async def test_zap_custom_error(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            raise ZAPError('some error', 408)

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                with pytest.raises(ZAPError) as error:
                    await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert error.value.text == 'some error'
    assert error.value.code == 408


@pytest.mark.asyncio
async def test_zap_successful_authentication_after_invalid_request(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            return 'bob', {b'foo': b'bar'}

    async with azmq.Context() as context:
        async with MyZAPAuthenticator(context=context):
            async with ZAPClient(context=context) as zap_client:
                async with context.socket(azmq.DEALER) as socket:
                    socket.connect(ZAP_INPROC_ENDPOINT)
                    await socket.send_multipart([b'invalid', b'data'])

                    username, metadata = await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert username == 'bob'
    assert metadata == {b'foo': b'bar'}


@pytest.mark.asyncio
async def test_zap_invalid_authenticator_response(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            return 'bob', {b'foo': b'bar'}

    async with azmq.Context() as context:
        async with ZAPClient(context=context) as zap_client:
            async with context.socket(azmq.ROUTER) as socket:
                socket.bind(ZAP_INPROC_ENDPOINT)

                task = asyncio.ensure_future(zap_client.authenticate(
                    domain='domain',
                    address='127.0.0.1',
                    identity=b'bob',
                    mechanism=b'CURVE',
                    credentials=[b'mycred', b'value'],
                ))
                identity, *_ = await asyncio.wait_for(
                    socket.recv_multipart(),
                    1,
                )
                await socket.send_multipart([
                    identity,
                    b'',
                    b'invalid',
                    b'message',
                ])

                assert not task.done()
                task.cancel()

            async with MyZAPAuthenticator(context=context):
                username, metadata = await asyncio.wait_for(
                    zap_client.authenticate(
                        domain='domain',
                        address='127.0.0.1',
                        identity=b'bob',
                        mechanism=b'CURVE',
                        credentials=[b'mycred', b'value'],
                    ),
                    1,
                )
    assert username == 'bob'
    assert metadata == {b'foo': b'bar'}


@pytest.mark.asyncio
async def test_zap_unknown_request(event_loop):
    class MyZAPAuthenticator(BaseZAPAuthenticator):
        async def on_request(self, *args, **kwargs):
            return 'bob', {b'foo': b'bar'}

    async with azmq.Context() as context:
        async with ZAPClient(context=context) as zap_client:
            async with context.socket(azmq.ROUTER) as socket:
                socket.bind(ZAP_INPROC_ENDPOINT)

                task = asyncio.ensure_future(zap_client.authenticate(
                    domain='domain',
                    address='127.0.0.1',
                    identity=b'bob',
                    mechanism=b'CURVE',
                    credentials=[b'mycred', b'value'],
                ))
                identity, *_ = await asyncio.wait_for(
                    socket.recv_multipart(),
                    1,
                )
                await socket.send_multipart([
                    identity,
                    b'',
                    b'1.0',
                    b'my_request',
                    b'200',
                    b'all good',
                    b'userid',
                    b'',
                ])

                assert not task.done()
                task.cancel()

            async with MyZAPAuthenticator(context=context):
                username, metadata = await asyncio.wait_for(
                    zap_client.authenticate(
                        domain='domain',
                        address='127.0.0.1',
                        identity=b'bob',
                        mechanism=b'CURVE',
                        credentials=[b'mycred', b'value'],
                    ),
                    1,
                )
    assert username == 'bob'
    assert metadata == {b'foo': b'bar'}


@pytest.mark.asyncio
async def test_zap_default_authenticator_authentication_failure(event_loop):
    async with azmq.Context() as context:
        async with ZAPAuthenticator(context=context) as authenticator:
            authenticator.allow('192.168.0.1')

            async with ZAPClient(context=context) as zap_client:
                with pytest.raises(ZAPAuthenticationFailure) as error:
                    await asyncio.wait_for(
                        zap_client.authenticate(
                            domain='domain',
                            address='127.0.0.1',
                            identity=b'bob',
                            mechanism=b'CURVE',
                            credentials=[b'mycred', b'value'],
                        ),
                        1,
                    )

    assert error.value.code == 400


@pytest.mark.asyncio
async def test_zap_default_authenticator_authentication_success(event_loop):
    async with azmq.Context() as context:
        async with ZAPAuthenticator(context=context) as authenticator:
            authenticator.deny('192.168.0.1')

            async with ZAPClient(context=context) as zap_client:
                username, metadata = await asyncio.wait_for(
                    zap_client.authenticate(
                        domain='domain',
                        address='127.0.0.1',
                        identity=b'bob',
                        mechanism=b'NULL',
                        credentials=[],
                    ),
                    1,
                )

    assert username == ''
    assert metadata == {}



"""
The connection mechanisms.
"""

from .curve_client import CurveClient
from .curve_server import CurveServer
from .null import Null
from .plain_client import PlainClient
from .plain_server import PlainServer

__all__ = [
    'CurveClient',
    'CurveServer',
    'Null',
    'PlainClient',
    'PlainServer',
]



"""
Unit tests for containers.
"""

import asyncio
import pytest

from azmq.containers import AsyncList


async def assert_empty(l):
    await asyncio.wait_for(l.wait_empty(), 0)

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(l.wait_not_empty(), 0)


async def assert_non_empty(l):
    await asyncio.wait_for(l.wait_not_empty(), 0)

    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(l.wait_empty(), 0)


def test_asynclist_list_likeness():
    al = AsyncList()
    al.append(4)
    al.extend([5, 6])
    assert al == [4, 5, 6]
    assert len(al) == 3
    assert al == AsyncList([4, 5, 6])


@pytest.mark.asyncio
async def test_asynclist_wait(event_loop):
    al = AsyncList()
    await assert_empty(al)

    al.append(5)
    await assert_non_empty(al)

    al.remove(5)
    await assert_empty(al)

    al.extend([4, 6])
    await assert_non_empty(al)

    al.remove(4)
    await assert_non_empty(al)

    al.clear()
    await assert_empty(al)

    al[:] = [1]
    await assert_non_empty(al)

    al[:] = []
    await assert_empty(al)

    al.insert(0, 4)
    await assert_non_empty(al)

    al.pop(0)
    await assert_empty(al)

    # call_soon() would run the call right away since we are in a coroutine.
    # This is not documented but it seems to happen consistently.
    event_loop.call_later(0, al.append, 2)
    await asyncio.wait_for(al.wait_change(), 0.5)


@pytest.mark.asyncio
async def test_fairlistproxy(event_loop):
    al = AsyncList()
    flp = al.create_proxy()
    assert not flp
    assert len(flp) == 0
    await assert_empty(flp)

    al.extend(['a', 'b', 'c', 'd'])
    assert flp
    assert len(flp) == len(al)
    await assert_non_empty(flp)

    assert list(flp) == ['a', 'b', 'c', 'd']
    assert list(flp) == ['b', 'c', 'd', 'a']
    al.remove('c')
    assert list(flp) == ['d', 'a', 'b']
    al.remove('d')
    assert list(flp) == ['a', 'b']
    al.remove('b')
    assert list(flp) == ['a']
    al.remove('a')
    assert list(flp) == []

    al.extend(['a', 'b', 'c'])
    assert next(iter(flp)) == 'a'
    assert next(iter(flp)) == 'b'
    assert next(iter(flp)) == 'c'
    assert next(iter(flp)) == 'a'

    # call_soon() would run the call right away since we are in a coroutine.
    # This is not documented but it seems to happen consistently.
    event_loop.call_later(0, al.append, 2)
    await asyncio.wait_for(flp.wait_change(), 0.5)



"""
Unit tests for the multiplexer.
"""

import pytest

from mock import MagicMock

from azmq.multiplexer import Multiplexer


@pytest.mark.asyncio
async def test_add_socket_twice(event_loop):
    socket = MagicMock()
    multiplexer = Multiplexer(loop=event_loop)
    multiplexer.add_socket(socket)
    multiplexer.add_socket(socket)

    assert multiplexer._sockets == {socket}


@pytest.mark.asyncio
async def test_remove_unadded_socket(event_loop):
    socket = MagicMock()
    multiplexer = Multiplexer(loop=event_loop)
    multiplexer.remove_socket(socket)

    assert multiplexer._sockets == set()


@pytest.mark.asyncio
async def test_recv_multipart_no_sockets(event_loop):
    multiplexer = Multiplexer(loop=event_loop)
    result = await multiplexer.recv_multipart()

    assert result == []



"""
A ZMQ socket class implementation.
"""

import asyncio
import random
import struct

from urllib.parse import (
    urlsplit,
    urlunsplit,
)
from contextlib import ExitStack
from functools import partial

from .common import (
    CompositeClosableAsyncObject,
    cancel_on_closing,
    AsyncBox,
)
from .constants import (
    DEALER,
    PAIR,
    PUB,
    PUSH,
    PULL,
    REP,
    REQ,
    ROUTER,
    SUB,
    XPUB,
    XSUB,
)
from .containers import AsyncList
from .errors import (
    UnsupportedSchemeError,
    InvalidOperation,
)
from .engines.tcp import (
    TCPClientEngine,
    TCPServerEngine,
)
from .engines.ipc import (
    IPCClientEngine,
    IPCServerEngine,
)
from .engines.inproc import (
    InprocClientEngine,
    InprocServerEngine,
)
from .log import logger
from .mechanisms import Null


class Peer(object):
    """
    Represents a peer connection, either establish or in the process of being
    established.
    """
    __slots__ = [
        'engine',
        'connection',
        'inbox',
        'outbox',
    ]

    def __init__(self, engine, connection, inbox, outbox):
        self.engine = engine
        self.connection = connection
        self.inbox = inbox
        self.outbox = outbox

    def reset(self):
        self.connection = None
        self.inbox.close()
        self.inbox = self.inbox.clone()
        self.outbox.close()
        self.outbox = self.outbox.clone()

    @property
    def dead(self):
        return not self.engine and self.inbox.empty()


class Socket(CompositeClosableAsyncObject):
    """
    A ZMQ socket.

    This class is **NOT** thread-safe.
    """
    def on_open(self, context, socket_type, identity=None, mechanism=None):
        super().on_open()

        self.context = context
        self.socket_type = socket_type
        self.identity = identity
        self.mechanism = mechanism or Null()
        self.max_inbox_size = 0
        self.max_outbox_size = 0
        self._outgoing_engines = {}
        self._outgoing_peers = {}
        self._incoming_engines = {}
        self._peers = AsyncList(loop=self.loop)
        self._in_peers = self._peers.create_proxy()
        self._out_peers = self._peers.create_proxy()
        self._base_identity = random.getrandbits(32)
        self._subscriptions = []
        self._read_lock = asyncio.Lock(loop=self.loop)

        if self.socket_type == REQ:
            # This future holds the last connection we sent a request to (or
            # none, if no request was sent yet). This allows to start receiving
            # before we send.
            self._current_peer = asyncio.Future(loop=self.loop)
            self.recv_multipart = self._recv_req
            self.send_multipart = self._send_req
        elif self.socket_type == REP:
            # This future holds the last connection we received a request from
            # (or none, if no request was received yet). This allows to start
            # receiving before we send.
            self._current_peer = asyncio.Future(loop=self.loop)
            self.recv_multipart = self._recv_rep
            self.send_multipart = self._send_rep
        elif self.socket_type == DEALER:
            self.recv_multipart = self._recv_dealer
            self.send_multipart = self._send_dealer
        elif self.socket_type == ROUTER:
            self.recv_multipart = self._recv_router
            self.send_multipart = self._send_router
        elif self.socket_type == PUB:
            self.recv_multipart = self._no_recv
            self.send_multipart = self._send_pub
        elif self.socket_type == XPUB:
            self.recv_multipart = self._recv_xpub
            self.send_multipart = self._send_pub  # This is not a typo.
        elif self.socket_type == SUB:
            self.recv_multipart = self._recv_sub
            self.send_multipart = self._no_send
        elif self.socket_type == XSUB:
            self.recv_multipart = self._recv_sub  # This is not a typo.
            self.send_multipart = self._send_xsub
        elif self.socket_type == PUSH:
            self.recv_multipart = self._no_recv
            self.send_multipart = self._send_push
        elif self.socket_type == PULL:
            self.recv_multipart = self._recv_pull
            self.send_multipart = self._no_send
        elif self.socket_type == PAIR:
            self.recv_multipart = self._recv_pair
            self.send_multipart = self._send_pair
        else:
            raise ValueError("Unsupported socket type: %r" % self.socket_type)

    def create_inbox(self):
        return AsyncBox(
            maxsize=self.max_inbox_size,
            loop=self.loop,
        )

    def create_outbox(self):
        return AsyncBox(
            maxsize=self.max_outbox_size,
            loop=self.loop,
        )

    def connect(self, endpoint, *, on_connection_failure=None):
        url = urlsplit(endpoint)

        if url.scheme == 'tcp':
            engine = TCPClientEngine(
                host=url.hostname,
                port=url.port,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        elif url.scheme == 'ipc':
            engine = IPCClientEngine(
                path=url.path,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        elif url.scheme == 'inproc':
            engine = InprocClientEngine(
                context=self.context,
                path=url.netloc,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        else:
            raise UnsupportedSchemeError(scheme=url.scheme)

        engine.on_connection_ready.connect(
            partial(self.register_connection, engine=engine),
        )
        engine.on_connection_lost.connect(
            partial(self.unregister_connection, engine=engine),
        )

        if on_connection_failure:
            engine.on_connection_failure.connect(on_connection_failure)

        self._outgoing_engines[url] = engine
        peer = Peer(
            engine=engine,
            connection=None,
            inbox=self.create_inbox(),
            outbox=self.create_outbox(),
        )
        self._outgoing_peers[engine] = peer
        self._peers.append(peer)
        self.register_child(engine)

    async def disconnect(self, endpoint):
        url = urlsplit(endpoint)

        engine = self._outgoing_engines.pop(url)
        peer = self._outgoing_peers.pop(engine)
        self._peers.remove(peer)
        engine.close()
        await engine.wait_closed()

    async def reconnect(self, endpoint, *, on_connection_failure=None):
        await self.disconnect(endpoint)
        self.connect(endpoint, on_connection_failure=on_connection_failure)

    async def reconnect_all(self, *, on_connection_failure=None):
        for endpoint in list(self._outgoing_engines):
            await self.reconnect(
                urlunsplit(endpoint),
                on_connection_failure=on_connection_failure,
            )

    def bind(self, endpoint, *, on_connection_failure=None):
        url = urlsplit(endpoint)

        if url.scheme == 'tcp':
            engine = TCPServerEngine(
                host=url.hostname,
                port=url.port,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        elif url.scheme == 'ipc':
            engine = IPCServerEngine(
                path=url.path,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        elif url.scheme == 'inproc':
            engine = InprocServerEngine(
                context=self.context,
                path=url.netloc,
                zap_client=self.context.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                loop=self.loop,
            )
        else:
            raise UnsupportedSchemeError(scheme=url.scheme)

        engine.on_connection_ready.connect(
            partial(self.register_connection, engine=engine),
        )
        engine.on_connection_lost.connect(
            partial(self.unregister_connection, engine=engine),
        )

        if on_connection_failure:
            engine.on_connection_failure.connect(on_connection_failure)

        self._incoming_engines[url] = engine
        self.register_child(engine)

    async def unbind(self, endpoint):
        url = urlsplit(endpoint)

        engine = self._incoming_engines.pop(url)
        engine.close()
        await engine.wait_closed()

    async def rebind(self, endpoint, *, on_connection_failure=None):
        await self.unbind(endpoint)
        self.bind(endpoint, on_connection_failure=on_connection_failure)

    async def rebind_all(self, *, on_connection_failure=None):
        for endpoint in list(self._incoming_engines):
            await self.rebind(
                urlunsplit(endpoint),
                on_connection_failure=on_connection_failure,
            )

    async def reset_all(self, *, on_connection_failure=None):
        await self.reconnect_all(on_connection_failure=on_connection_failure)
        await self.rebind_all(on_connection_failure=on_connection_failure)

    def register_connection(self, connection, engine):
        logger.debug("Registering new active connection: %s", connection)

        peer = self._outgoing_peers.get(engine)

        if peer:
            peer.connection = connection
            logger.debug("Updating connected peer with it's new connection.")
        else:
            peer = Peer(
                engine=engine,
                connection=connection,
                inbox=self.create_inbox(),
                outbox=self.create_outbox(),
            )
            self._peers.append(peer)
            logger.debug("Creating a new peer with the new connection.")

        connection.set_queues(inbox=peer.inbox, outbox=peer.outbox)

        if self.socket_type == ROUTER and not connection.remote_identity:
            connection.remote_identity = self.generate_identity()
            logger.info(
                "Peer did not specify an identity. Generated one for him "
                "(%r).",
                connection.remote_identity,
            )

        if self.socket_type in {SUB, XSUB}:
            for topic in self._subscriptions:
                asyncio.ensure_future(
                    connection.local_subscribe(topic),
                    loop=self.loop,
                )

    def unregister_connection(self, connection, engine):
        logger.debug("Unregistering inactive connection: %s", connection)

        peer = self._outgoing_peers.get(engine)

        if not peer:
            # If the peer is not an outgoing peer, we remove its associated
            # engine so it gets removed when its inbox becomes empty.
            peer = next(  # pragma: no cover
                p for p in self._peers if p.connection is connection,
            )
            peer.engine = None

        # This cancels any pending read on the peer.
        peer.reset()

    def generate_identity(self):
        """
        Generate a unique but random identity.
        """
        identity = struct.pack('!BI', 0, self._base_identity)
        self._base_identity += 1

        if self._base_identity >= 2 ** 32:
            self._base_identity = 0

        return identity

    async def _wait_peers(self):
        """
        Blocks until at least one non-dead peer is available.
        """
        # Make sure we remove dead peers.
        for p in self._peers[:]:
            if p.dead:
                self._peers.remove(p)

        while not self._peers:
            await self._peers.wait_not_empty()

    async def _fair_get_in_peer(self):
        """
        Get the first available available inbound peer in a fair manner.

        :returns: A `Peer` inbox, whose inbox is guaranteed not to be
            empty (and thus can be read from without blocking).
        """
        peer = None

        while not peer:
            await self._wait_peers()

            # This rotates the list, implementing fair-queuing.
            peers = list(self._in_peers)

            tasks = [asyncio.ensure_future(self._in_peers.wait_change())]
            tasks.extend([
                asyncio.ensure_future(
                    p.inbox.wait_not_empty(),
                    loop=self.loop,
                )
                for p in peers
            ])

            try:
                done, pending = await asyncio.wait(
                    tasks,
                    return_when=asyncio.FIRST_COMPLETED,
                    loop=self.loop,
                )
            finally:
                for task in tasks:
                    task.cancel()

            tasks.pop(0)  # pop the wait_change task.
            peer = next(
                (
                    p
                    for task, p in zip(tasks, peers)
                    if task in done and not task.cancelled()
                ),
                None,
            )

        return peer

    async def _fair_recv(self):
        """
        Receive from all the existing peers, rotating the list of peers every
        time.

        :returns: The frames.
        """
        with await self._read_lock:
            peer = await self._fair_get_in_peer()
            result = peer.inbox.read_nowait()

        return result

    async def _fair_get_out_peer(self):
        """
        Get the first available peer, with non-blocking inbox or wait until one
        meets the condition.

        :returns: The peer whose outbox is ready to be written to.
        """
        peer = None

        while not peer:
            await self._wait_peers()

            # This rotates the list, implementing fair-queuing.
            peers = list(self._out_peers)

            tasks = [asyncio.ensure_future(self._out_peers.wait_change())]
            tasks.extend([
                asyncio.ensure_future(
                    p.outbox.wait_not_full(),
                    loop=self.loop,
                )
                for p in peers
            ])

            try:
                done, pending = await asyncio.wait(
                    tasks,
                    return_when=asyncio.FIRST_COMPLETED,
                    loop=self.loop,
                )
            finally:
                for task in tasks:
                    task.cancel()

            tasks.pop(0)  # pop the wait_change task.
            peer = next(  # pragma: no cover
                (
                    p
                    for task, p in zip(tasks, peers)
                    if task in done and not p.outbox.full()
                ),
                None,
            )

        return peer

    async def _fair_send(self, frames):
        """
        Send from the first available, non-blocking peer or wait until one
        meets the condition.

        :params frames: The frames to write.
        :returns: The peer that was used.
        """
        peer = await self._fair_get_out_peer()
        peer.outbox.write_nowait(frames)
        return peer

    @cancel_on_closing
    async def _send_req(self, frames):
        if self._current_peer.done():
            raise InvalidOperation(
                "Cannot send twice in a row from a REQ socket. Please recv "
                "from it first",
            )

        peer = await self._fair_send([b''] + frames)
        self._current_peer.set_result(peer)

    @cancel_on_closing
    async def _recv_req(self):
        await self._current_peer
        peer = self._current_peer.result()

        # Let's allow writes back as soon as we know on which peer to receive.
        self._current_peer = asyncio.Future(loop=self.loop)

        # As per 28/REQREP, a REQ socket SHALL discard silently any messages
        # received from other peers when processing incoming messages on the
        # current connection.
        with ExitStack() as stack:
            for p in self._peers:
                if p is not peer:
                    stack.enter_context(
                        p.connection.discard_incoming_messages(),
                    )

            frames = await peer.inbox.read()

            # We need to get rid of the empty delimiter.
            if frames[0] != b'':
                logger.warning(
                    "Received unexpected reply (%r) in REQ socket. Closing "
                    "connection. The recv() call will NEVER return !",
                    frames,
                )
                peer.connection.close()

                # This may seem weird but we must treat these errors as if the
                # peer did never reply, which means blocking forever (at least
                # until the socket is closed).
                #
                # ZMQ best-practices dictate the user should recreate the
                # socket anyway in case of timeouts and there is no other
                # sensible course of action: we can't return anything
                # meaningful and throwing an error puts the burden on the user
                # by forcing him/her to handle two possible outcomes.
                forever = asyncio.Future(loop=self.loop)
                await forever

            frames.pop(0)
            return frames

    @cancel_on_closing
    async def _send_rep(self, frames):
        await self._current_peer
        peer, envelope = self._current_peer.result()

        # Let's allow reads back as soon as we know on which connection to
        # receive.
        self._current_peer = asyncio.Future(loop=self.loop)

        await peer.outbox.write(envelope + frames)

    @cancel_on_closing
    async def _recv_rep(self):
        if self._current_peer.done():
            raise InvalidOperation(
                "Cannot receive twice in a row from a REP socket. Please send "
                "from it first",
            )

        peer = await self._fair_get_in_peer()
        frames = peer.inbox.read_nowait()
        delimiter_index = frames.index(b'')
        envelope = frames[:delimiter_index + 1]
        message = frames[delimiter_index + 1:]
        self._current_peer.set_result((peer, envelope))
        return message

    @cancel_on_closing
    async def _send_dealer(self, frames):
        await self._fair_send(frames)

    @cancel_on_closing
    async def _recv_dealer(self):
        return await self._fair_recv()

    @cancel_on_closing
    async def _send_router(self, frames):
        identity = frames.pop(0)

        try:
            peer = next(
                p for p in self._peers
                if p.connection and
                p.connection.remote_identity == identity and
                not p.outbox.full()
            )
        except StopIteration:
            # We drop the messages as their is no suitable connection to write
            # it to.
            pass
        else:
            peer.outbox.write_nowait(frames)

    @cancel_on_closing
    async def _recv_router(self):
        peer = None

        with await self._read_lock:
            while not peer or not peer.connection:
                peer = await self._fair_get_in_peer()

        frames = peer.inbox.read_nowait()
        frames.insert(0, peer.connection.remote_identity)
        return frames

    @cancel_on_closing
    async def _no_recv(self):
        raise AssertionError(
            "A %s socket cannot receive." % self.socket_type.decode(),
        )

    @cancel_on_closing
    async def _no_send(self, *args):
        raise AssertionError(
            "A %s socket cannot send." % self.socket_type.decode(),
        )

    @cancel_on_closing
    async def _send_pub(self, frames):
        topic = frames[0]

        for peer in self._peers:
            if peer.connection and not peer.outbox.full() and next(
                (
                    topic.startswith(subs)
                    for subs in peer.connection.subscriptions
                ),
                None,
            ):
                peer.outbox.write_nowait(frames)

    @cancel_on_closing
    async def _recv_xpub(self):
        return await self._fair_recv()

    @cancel_on_closing
    async def _recv_sub(self):
        return await self._fair_recv()

    @cancel_on_closing
    async def _send_xsub(self, frames):
        if frames and frames[0]:
            first_frame = frames[0]
            type_ = first_frame[0]

            if type_ == 0:
                await self.unsubscribe(first_frame[1:])
            elif type_ == 1:
                await self.subscribe(first_frame[1:])

    @cancel_on_closing
    async def _send_push(self, frames):
        await self._fair_send(frames)

    @cancel_on_closing
    async def _recv_pull(self):
        return await self._fair_recv()

    @cancel_on_closing
    async def _send_pair(self, frames):
        # We only send to the first connected peer.
        peer = None

        while not peer:
            await self._peers.wait_not_empty()
            peer = self._peers[0]

            try:
                await peer.outbox.write(frames)
            except asyncio.CancelledError:  # pragma: no cover
                # This is impossible to test reliably.
                peer = None

    @cancel_on_closing
    async def _recv_pair(self):
        # We only receive from the first connected peer.
        peer = None

        while not peer:  # pragma: no branch
            await self._peers.wait_not_empty()
            peer = self._peers[0]

            try:
                return await peer.inbox.read()
            except asyncio.CancelledError:  # pragma: no cover
                # This is impossible to test reliably.
                peer = None

    @cancel_on_closing
    async def subscribe(self, topic):
        """
        Subscribe the socket to the specified topic.

        :param topic: The topic to subscribe to.
        """
        if self.socket_type not in {SUB, XSUB}:
            raise AssertionError(
                "A %s socket cannot subscribe." % self.socket_type.decode(),
            )

        # Do this **BEFORE** awaiting so that new connections created during
        # the execution below honor the setting.
        self._subscriptions.append(topic)
        tasks = [
            asyncio.ensure_future(
                peer.connection.local_subscribe(topic),
                loop=self.loop,
            )
            for peer in self._peers
            if peer.connection
        ]

        if tasks:
            try:
                await asyncio.wait(tasks, loop=self.loop)
            finally:
                for task in tasks:
                    task.cancel()

    @cancel_on_closing
    async def unsubscribe(self, topic):
        """
        Unsubscribe the socket from the specified topic.

        :param topic: The topic to unsubscribe from.
        """
        if self.socket_type not in {SUB, XSUB}:
            raise AssertionError(
                "A %s socket cannot unsubscribe." % self.socket_type.decode(),
            )

        # Do this **BEFORE** awaiting so that new connections created during
        # the execution below honor the setting.
        self._subscriptions.remove(topic)
        tasks = [
            asyncio.ensure_future(
                peer.connection.local_unsubscribe(topic),
                loop=self.loop,
            )
            for peer in self._peers
            if peer.connection
        ]

        if tasks:
            try:
                await asyncio.wait(tasks, loop=self.loop)
            finally:
                for task in tasks:
                    task.cancel()



"""
Unit tests for ZAP.
"""

import pytest

from azmq.context import Context
from azmq.zap import ZAPAuthenticator


@pytest.yield_fixture
def context(event_loop):
    context = Context(loop=event_loop)
    yield context
    context.close()


@pytest.yield_fixture
def authenticator(event_loop, context):
    authenticator = ZAPAuthenticator(
        context=context,
        loop=event_loop,
    )
    yield authenticator
    authenticator.close()


def test_zap_authenticator_allow(authenticator):
    authenticator.allow('127.0.0.1')
    assert authenticator.whitelist == ['127.0.0.1']
    assert authenticator.blacklist == []


def test_zap_authenticator_deny(authenticator):
    authenticator.deny('127.0.0.1')
    assert authenticator.blacklist == ['127.0.0.1']
    assert authenticator.whitelist == []


def test_zap_authenticator_deny_after_allow(authenticator):
    authenticator.allow('127.0.0.1')
    authenticator.deny('127.0.0.1')
    assert authenticator.blacklist == ['127.0.0.1']
    assert authenticator.whitelist == []


def test_zap_authenticator_allow_after_deny(authenticator):
    authenticator.deny('127.0.0.1')
    authenticator.allow('127.0.0.1')
    assert authenticator.whitelist == ['127.0.0.1']
    assert authenticator.blacklist == []


def test_zap_authenticator_add_user(authenticator):
    authenticator.add_user(username='bob', password='pwd')
    assert authenticator.passwords == {'bob': 'pwd'}


def test_zap_authenticator_remove_user(authenticator):
    authenticator.remove_user(username='bob')
    assert authenticator.passwords == {}


def test_zap_authenticator_remove_user_after_add(authenticator):
    authenticator.add_user(username='bob', password='pwd')
    authenticator.remove_user(username='bob')
    assert authenticator.passwords == {}


def test_zap_authenticator_add_authorized_key(authenticator):
    authenticator.add_authorized_key(b'mykey')
    assert authenticator.authorized_keys == {b'mykey'}


def test_zap_authenticator_remove_authorized_key(authenticator):
    authenticator.remove_authorized_key(b'mykey')
    assert authenticator.authorized_keys == set()


def test_zap_authenticator_remove_authorized_key_after_add(authenticator):
    authenticator.add_authorized_key(b'mykey')
    authenticator.remove_authorized_key(b'mykey')
    assert authenticator.authorized_keys == set()



"""
IPC engines.
"""

import asyncio
import os
import sys

from ..connections.stream import StreamConnection
from ..log import logger

from .base import BaseEngine

if sys.platform == 'win32':  # pragma: no cover
    from .win32 import start_pipe_server as start_ipc_server
    from .win32 import open_pipe_connection as open_ipc_connection
else:  # pragma: no cover
    from asyncio import start_unix_server

    async def start_ipc_server(*args, **kwargs):
        try:
            os.unlink(kwargs['path'])
        except FileNotFoundError:
            pass

        return await start_unix_server(*args, **kwargs)

    from asyncio import open_unix_connection as open_ipc_connection


class IPCClientEngine(BaseEngine):
    def on_open(self, *, path, **kwargs):
        super().on_open(**kwargs)

        self.path = path

    async def open_connection(self):
        reader, writer = await open_ipc_connection(
            path=self.path,
            loop=self.loop,
        )

        logger.debug("Connection to %s established.", self.path)

        try:
            async with StreamConnection(
                reader=reader,
                writer=writer,
                address=self.path,
                zap_client=self.zap_client,
                socket_type=self.socket_type,
                identity=self.identity,
                mechanism=self.mechanism,
                on_ready=self.on_connection_ready.emit,
                on_lost=self.on_connection_lost.emit,
                on_failure=self.on_connection_failure,
                loop=self.loop,
            ) as connection:
                self.register_child(connection)
                await connection.wait_closed()

            return connection.get_error()
        finally:
            logger.debug("Connection to %s closed.", self.path)


class IPCServerEngine(BaseEngine):
    def on_open(self, *, path, **kwargs):
        super().on_open(**kwargs)

        self.path = path

    async def open_connection(self):
        try:
            server = await start_ipc_server(
                self.handle_connection,
                path=self.path,
                loop=self.loop,
            )

            try:
                await server.wait_closed()
            finally:
                server.close()
                await server.wait_closed()

        except asyncio.CancelledError:
            raise
        except Exception:
            logger.exception(
                "Unable to start UNIX server on %s.",
                self.path,
            )

    async def handle_connection(self, reader, writer):
        logger.debug("Connection from %s established.", self.path)

        async with StreamConnection(
            reader=reader,
            writer=writer,
            address=self.path,
            zap_client=self.zap_client,
            socket_type=self.socket_type,
            identity=self.identity,
            mechanism=self.mechanism,
            on_ready=self.on_connection_ready.emit,
            on_lost=self.on_connection_lost.emit,
            on_failure=self.on_connection_failure,
            loop=self.loop,
        ) as connection:
            self.register_child(connection)
            await connection.wait_closed()

        logger.debug("Connection from %s lost.", self.path)



"""
Unit tests for contexts.
"""

import pytest

from mock import MagicMock

from azmq.context import Context
from azmq.errors import InprocPathBound


def test_set_zap_authenticator(event_loop):
    context = Context(loop=event_loop)
    zap_authenticator = MagicMock()
    context.set_zap_authenticator(zap_authenticator)
    assert context._zap_authenticator is zap_authenticator
    result = context.set_zap_authenticator(None)
    assert result is zap_authenticator
    assert context._zap_authenticator is None


@pytest.mark.asyncio
async def test_start_inproc_server_twice(event_loop):
    context = Context(loop=event_loop)
    handler = MagicMock()
    await context._start_inproc_server(handler, 'boo')

    with pytest.raises(InprocPathBound) as error:
        await context._start_inproc_server(handler, 'boo')

    assert error.value.path == 'boo'



"""
ZMQ context class implementation.
"""

import asyncio

from .common import (
    CompositeClosableAsyncObject,
    cancel_on_closing,
)
from .log import logger
from .errors import InprocPathBound
from .socket import Socket
from .transports.inproc import InprocServer
from .zap import ZAPClient


class Context(CompositeClosableAsyncObject):
    """
    A ZMQ context.

    This class is **NOT** thread-safe.
    """
    def on_open(self):
        super().on_open()

        self._inproc_servers = {}
        self._zap_authenticator = None
        self.zap_client = None

    async def on_close(self):
        for server_future in self._inproc_servers.values():
            server_future.cancel()

        self._inproc_servers.clear()

        await super().on_close()

    def socket(self, socket_type, identity=None, mechanism=None):
        """
        Create and register a new socket.

        :param socket_type: The type of the socket.
        :param loop: An optional event loop to associate the socket with.

        This is the preferred method to create new sockets.
        """
        socket = Socket(
            context=self,
            socket_type=socket_type,
            identity=identity,
            mechanism=mechanism,
            loop=self.loop,
        )
        self.register_child(socket)
        return socket

    def set_zap_authenticator(self, zap_authenticator):
        """
        Setup a ZAP authenticator.

        :param zap_authenticator: A ZAP authenticator instance to use. The
            context takes ownership of the specified instance. It will close it
            automatically when it stops. If `None` is specified, any previously
            owner instance is disowned and returned. It becomes the caller's
            responsibility to close it.
        :returns: The previous ZAP authenticator instance.
        """
        result = self._zap_authenticator

        if result:
            self.unregister_child(result)

        self._zap_authenticator = zap_authenticator

        if self.zap_client:
            self.zap_client.close()

        if self._zap_authenticator:
            self.register_child(zap_authenticator)
            self.zap_client = ZAPClient(context=self)
            self.register_child(self.zap_client)
        else:
            self.zap_client = None

        return result

    @cancel_on_closing
    async def _start_inproc_server(self, handler, path):
        server_future = self._inproc_servers.get(path)

        if not server_future:
            server_future = self._inproc_servers[path] = \
                asyncio.Future(loop=self.loop)
        elif server_future.done():
            server = server_future.result()

            if server.closing:
                await server.wait_closed()
                server_future = self._inproc_servers[path] = \
                    asyncio.Future(loop=self.loop)
            else:
                raise InprocPathBound(path=path)

        server = InprocServer(loop=self.loop, handler=handler)
        logger.debug("Starting inproc server on %r.", path)
        self.register_child(server)
        server_future.set_result(server)
        return server

    @cancel_on_closing
    async def _open_inproc_connection(self, path):
        server_future = self._inproc_servers.get(path)
        server = None

        while not server or server.closing:
            if not server_future or server_future.cancelled():
                server_future = self._inproc_servers[path] = \
                    asyncio.Future(loop=self.loop)

            await asyncio.shield(server_future)
            server = server_future.result()
            server_future = None

        return await server.create_channel(path)



"""
A ZMTP connection over the inproc transport.
"""

import asyncio

from ..constants import (
    PUB,
    XPUB,
)
from ..errors import ProtocolError
from ..log import logger
from .base import BaseConnection


class InprocConnection(BaseConnection):
    """
    Implements a ZMTP connection that works in-process.
    """
    def __init__(
        self,
        channel,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.channel = channel
        self.register_child(self.channel)

    async def on_close(self):
        self.channel.close()
        return await super().on_close()

    async def on_run(self):
        await self.channel.write(self.get_metadata())
        metadata = await self.channel.read()
        self.set_remote_metadata(metadata)

        logger.debug("Connection is now ready to read and write.")
        self.on_ready(self)

        assert self.inbox or self.outbox, (
            "on_ready callback must either set an inbox or an outbox."
        )

        try:
            tasks = []
            read_task = asyncio.ensure_future(self.read(), loop=self.loop)
            tasks.append(read_task)
            write_task = asyncio.ensure_future(
                self.write(),
                loop=self.loop,
            )
            tasks.append(write_task)

            try:
                await self.await_until_closing(
                    asyncio.gather(*tasks, loop=self.loop),
                )
            except asyncio.CancelledError:
                logger.debug(
                    "Read/write cycle interrupted. Connection will die soon.",
                )
            finally:
                for task in tasks:
                    task.cancel()

                await asyncio.wait(tasks, loop=self.loop)
        finally:
            self.on_lost(self)

        # Flush out the unset outgoing messages before we exit.
        while not self.outbox.empty():
            await self.channel.write(self.outbox.read_nowait())

    async def read(self):
        while True:
            frames = await self.channel.read()

            if self.socket_type in {PUB, XPUB}:
                type_ = frames[0][0]

                if type_ == 1:
                    await self.subscribe(frames[0][1:])
                elif type_ == 0:
                    await self.unsubscribe(frames[0][1:])
                else:
                    raise ProtocolError(
                        "Unexpected subscription message type (%s)." % type_,
                    )
            elif not self._discard_incoming_messages:
                await self.inbox.write(frames)

    async def write(self):
        outbox = self.outbox

        while True:
            frames = await outbox.read()
            await self.channel.write(frames)



"""
Inproc integration tests.
"""

import asyncio
import pytest

import azmq

from azmq.multiplexer import Multiplexer
from azmq.errors import (
    InvalidOperation,
    ProtocolError,
    UnsupportedSchemeError,
)


from ..conftest import use_all_transports


async def zerosec(awaitable):
    """
    Assert that an awaitable would block.

    :param awaitable: The awaitable to wrap.
    :returns: A decorated awaitable.
    """
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(awaitable, 0.25)


def fivesec(awaitable):
    """
    Causes a normally blocking call to timeout after a while.

    :param awaitable: The awaitable to wrap.
    :returns: A decorated awaitable that times out.
    """
    return asyncio.wait_for(awaitable, 5)


@use_all_transports
@pytest.mark.asyncio
async def test_req_rep(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            req_socket.bind(endpoint)
            rep_socket.connect(endpoint)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            message = await fivesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']
            await fivesec(rep_socket.send_multipart([b'my', b'response']))
            message = await req_socket.recv_multipart()
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_push_pull(event_loop, endpoint):
    async with azmq.Context() as context:
        push_socket = context.socket(azmq.PUSH)
        pull_socket_1 = context.socket(azmq.PULL)
        pull_socket_2 = context.socket(azmq.PULL)

        try:
            push_socket.bind(endpoint)
            pull_socket_1.connect(endpoint)
            pull_socket_2.connect(endpoint)

            multiplexer = Multiplexer()
            multiplexer.add_socket(pull_socket_1)
            multiplexer.add_socket(pull_socket_2)

            await fivesec(push_socket.send_multipart([b'a', b'1']))
            await fivesec(push_socket.send_multipart([b'b', b'2']))

            messages = []

            while len(messages) < 2:
                results = await fivesec(multiplexer.recv_multipart())
                messages.extend(tuple(x) for _, x in results)

            assert set(messages) == {(b'a', b'1'), (b'b', b'2')}

        finally:
            push_socket.close()
            pull_socket_1.close()
            pull_socket_2.close()


@use_all_transports
@pytest.mark.asyncio
async def test_push_pull_slow_bind(event_loop, endpoint):
    async with azmq.Context() as context:
        push_socket = context.socket(azmq.PUSH)
        pull_socket = context.socket(azmq.PULL)

        try:
            # The PUSH socket connects before the PULL sockets binds and sends
            # a message right away.
            push_socket.connect(endpoint)
            await fivesec(push_socket.send_multipart([b'hello']))

            # The PULL sockets finally binds, and should receive the message,
            # even late.
            pull_socket.bind(endpoint)
            message = await fivesec(pull_socket.recv_multipart())
            assert message == [b'hello']

        finally:
            push_socket.close()
            pull_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_push_pull_slow_connect(event_loop, endpoint):
    async with azmq.Context() as context:
        push_socket = context.socket(azmq.PUSH)
        pull_socket = context.socket(azmq.PULL)

        try:
            # The PUSH socket binds before the PULL sockets binds and sends
            # a message right away. It should block in this case.
            push_socket.bind(endpoint)
            await zerosec(push_socket.send_multipart([b'hello']))

            # The PULL sockets finally binds, and should receive no message.
            pull_socket.connect(endpoint)
            await zerosec(pull_socket.recv_multipart())

        finally:
            push_socket.close()
            pull_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_push_pull_explicit_reconnect(event_loop, endpoint):
    async with azmq.Context() as context:
        push_socket = context.socket(azmq.PUSH)
        pull_socket = context.socket(azmq.PULL)

        try:
            push_socket.connect(endpoint)
            await fivesec(push_socket.send_multipart([b'hello']))

            # The disconnection should cause the outgoing message to be lost.
            await fivesec(push_socket.reconnect(endpoint))

            pull_socket.bind(endpoint)
            await zerosec(pull_socket.recv_multipart())

            await fivesec(push_socket.send_multipart([b'hello']))
            message = await fivesec(pull_socket.recv_multipart())
            assert message == [b'hello']

        finally:
            push_socket.close()
            pull_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_implicit_reconnect(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_socket = context.socket(azmq.SUB)
        await sub_socket.subscribe(b'h')

        try:
            pub_socket.bind(endpoint)
            sub_socket.connect(endpoint)

            # Let's start a task that sends messages on the pub socket.
            async def publish():
                while True:
                    await fivesec(pub_socket.send_multipart([b'hello']))

            publish_task = asyncio.ensure_future(publish())

            try:
                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']

                await fivesec(pub_socket.rebind(endpoint))

                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']
            finally:
                publish_task.cancel()

        finally:
            pub_socket.close()
            sub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_reset(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_socket = context.socket(azmq.SUB)
        await sub_socket.subscribe(b'h')

        try:
            pub_socket.bind(endpoint)
            sub_socket.connect(endpoint)

            # Let's start a task that sends messages on the pub socket.
            async def publish():
                while True:
                    await fivesec(pub_socket.send_multipart([b'hello']))

            publish_task = asyncio.ensure_future(publish())

            try:
                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']

                await fivesec(sub_socket.reset_all())
                await fivesec(pub_socket.reset_all())

                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']
            finally:
                publish_task.cancel()

        finally:
            pub_socket.close()
            sub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_spam(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_sockets = [
            context.socket(azmq.SUB)
            for _ in range(10)
        ]

        try:
            async def send():
                pub_socket.bind(endpoint)

                while True:
                    await fivesec(pub_socket.send_multipart([b'a', b'b']))

            async def recv(socket):
                await socket.subscribe(b'a')
                socket.connect(endpoint)
                message = await fivesec(socket.recv_multipart())
                assert message == [b'a', b'b']

            send_task = asyncio.ensure_future(send())
            recv_tasks = [
                asyncio.ensure_future(recv(socket))
                for socket in sub_sockets
            ]

            try:
                await fivesec(asyncio.gather(*recv_tasks))
            finally:
                send_task.cancel()

        finally:
            pub_socket.close()

            for socket in sub_sockets:
                socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_spam_subscribe_after(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_sockets = [
            context.socket(azmq.SUB)
            for _ in range(10)
        ]

        try:
            async def send():
                pub_socket.bind(endpoint)

                while True:
                    await fivesec(pub_socket.send_multipart([b'a', b'b']))

            async def recv(socket):
                socket.connect(endpoint)
                await socket.subscribe(b'a')
                message = await fivesec(socket.recv_multipart())
                assert message == [b'a', b'b']

            send_task = asyncio.ensure_future(send())
            recv_tasks = [
                asyncio.ensure_future(recv(socket))
                for socket in sub_sockets
            ]

            try:
                await fivesec(asyncio.gather(*recv_tasks))
            finally:
                send_task.cancel()

        finally:
            pub_socket.close()

            for socket in sub_sockets:
                socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_push_pull_max_outbox_size(event_loop, endpoint):
    async with azmq.Context() as context:
        push_socket = context.socket(azmq.PUSH)
        push_socket.max_outbox_size = 1
        pull_socket = context.socket(azmq.PULL)

        try:
            # The PUSH socket connects before the PULL sockets binds and sends
            # two messages right away. The first one should not block but the
            # second should as the queue size is 1 and the socket can't
            # possibly have sent the element so far.
            push_socket.connect(endpoint)
            await fivesec(push_socket.send_multipart([b'one']))
            await zerosec(push_socket.send_multipart([b'two']))

            # The PULL sockets finally binds, and should receive the first
            # message, even late.
            pull_socket.bind(endpoint)
            message = await fivesec(pull_socket.recv_multipart())
            assert message == [b'one']
            await zerosec(pull_socket.recv_multipart())

        finally:
            push_socket.close()
            pull_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_router_identity(event_loop, endpoint):
    async with azmq.Context() as context:
        dealer_socket = context.socket(azmq.DEALER)
        router_socket = context.socket(azmq.ROUTER)

        try:
            dealer_socket.identity = b'hello'
            dealer_socket.bind(endpoint)
            router_socket.connect(endpoint)

            await fivesec(dealer_socket.send_multipart(
                [b'', b'my', b'request'],
            ))
            message = await fivesec(router_socket.recv_multipart())
            assert message == [b'hello', b'', b'my', b'request']
            await fivesec(router_socket.send_multipart(
                [b'hello', b'', b'my', b'response'],
            ))
            message = await dealer_socket.recv_multipart()
            assert message == [b'', b'my', b'response']

        finally:
            dealer_socket.close()
            router_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_rep_invalid_identity(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            future = asyncio.Future(loop=event_loop)
            req_socket.connect(
                endpoint,
                on_connection_failure=future.set_result,
            )
            rep_socket.identity = b'\0invalid'
            rep_socket.bind(endpoint)

            await fivesec(future)
            assert isinstance(future.result().error, ProtocolError)
            assert future.result().error.fatal

        finally:
            req_socket.close()
            rep_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_req_invalid_combination(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        req_socket_2 = context.socket(azmq.REQ)

        try:
            future = asyncio.Future(loop=event_loop)
            req_socket.connect(
                endpoint,
                on_connection_failure=future.set_result,
            )
            req_socket_2.bind(endpoint)

            await fivesec(future)
            assert isinstance(future.result().error, ProtocolError)
            assert future.result().error.fatal

        finally:
            req_socket.close()
            req_socket_2.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_rep_multiple_receivers(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)
        rep_socket_2 = context.socket(azmq.REP)

        try:
            req_socket.bind(endpoint)
            rep_socket.connect(endpoint)
            rep_socket_2.connect(endpoint)

            multiplexer = Multiplexer()
            multiplexer.add_socket(rep_socket)
            multiplexer.add_socket(rep_socket_2)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            (socket, message), = await fivesec(multiplexer.recv_multipart())
            assert message == [b'my', b'request']
            await fivesec(socket.send_multipart([b'my', b'response']))
            message = await req_socket.recv_multipart()
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()
            rep_socket_2.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_router_multiple_receivers(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        router_socket = context.socket(azmq.ROUTER)
        router_socket_2 = context.socket(azmq.ROUTER)
        router_sockets = {router_socket, router_socket_2}

        try:
            req_socket.identity = b'me'
            req_socket.bind(endpoint)
            router_socket.connect(endpoint)
            router_socket_2.connect(endpoint)

            multiplexer = Multiplexer()
            multiplexer.add_socket(router_socket_2)
            multiplexer.add_socket(router_socket)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            (socket, message), = await fivesec(multiplexer.recv_multipart())
            assert message == [b'me', b'', b'my', b'request']
            other_socket = next(iter(router_sockets - {socket}))

            task = asyncio.ensure_future(req_socket.recv_multipart())

            await fivesec(other_socket.send_multipart(
                [b'me', b'', b'some', b'error']
            ))
            await fivesec(socket.send_multipart(
                [b'me', b'', b'my', b'response']
            ))
            message = await fivesec(task)
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            router_socket.close()
            router_socket_2.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_unsubscribe(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_socket = context.socket(azmq.SUB)
        await sub_socket.subscribe(b'h')
        await sub_socket.subscribe(b'g')

        try:
            pub_socket.bind(endpoint)
            sub_socket.connect(endpoint)

            # Let's start a task that sends messages on the pub socket.
            async def publish():
                while True:
                    await fivesec(pub_socket.send_multipart([b'hello']))

            publish_task = asyncio.ensure_future(publish())

            try:
                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']
            finally:
                publish_task.cancel()

            await sub_socket.unsubscribe(b'h')

            # Let's start a task that sends messages on the pub socket.
            async def publish():
                while True:
                    await fivesec(pub_socket.send_multipart([b'hello']))
                    await fivesec(pub_socket.send_multipart([b'gello']))

            publish_task = asyncio.ensure_future(publish())

            try:
                while message == [b'hello']:
                    message = await fivesec(sub_socket.recv_multipart())

                assert message == [b'gello']
            finally:
                publish_task.cancel()

        finally:
            pub_socket.close()
            sub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_xpub_sub_subscriptions(event_loop, endpoint):
    async with azmq.Context() as context:
        xpub_socket = context.socket(azmq.XPUB)
        sub_socket = context.socket(azmq.SUB)
        await sub_socket.subscribe(b'a')
        await sub_socket.unsubscribe(b'a')
        await sub_socket.subscribe(b'a')

        try:
            xpub_socket.bind(endpoint)
            sub_socket.connect(endpoint)

            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\1a']
            await sub_socket.unsubscribe(b'a')
            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\0a']

        finally:
            xpub_socket.close()
            sub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_xpub_xsub_subscriptions(event_loop, endpoint):
    async with azmq.Context() as context:
        xpub_socket = context.socket(azmq.XPUB)
        xsub_socket = context.socket(azmq.XSUB)

        try:
            xpub_socket.bind(endpoint)
            xsub_socket.connect(endpoint)

            await xsub_socket.send_multipart([b'\1a'])
            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\1a']
            await xsub_socket.send_multipart([b'\0a'])
            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\0a']

        finally:
            xpub_socket.close()
            xsub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_xpub_xsub_invalid_unsubscription(event_loop, endpoint):
    async with azmq.Context() as context:
        xpub_socket = context.socket(azmq.XPUB)
        xsub_socket = context.socket(azmq.XSUB)

        try:
            xpub_socket.bind(endpoint)
            xsub_socket.connect(endpoint)

            # We force a protocol error by sending an unexpected message.
            await xsub_socket._fair_send([b'\0b'])
            await xsub_socket.send_multipart([b'\1a'])
            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\1a']

        finally:
            xpub_socket.close()
            xsub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_xpub_xsub_unknown_subscription_message(event_loop, endpoint):
    async with azmq.Context() as context:
        xpub_socket = context.socket(azmq.XPUB)
        xsub_socket = context.socket(azmq.XSUB)

        try:
            xpub_socket.bind(endpoint)
            xsub_socket.connect(endpoint)

            # We force a protocol error by sending an unexpected message.
            await xsub_socket._fair_send([b'\2x'])
            await xsub_socket.send_multipart([b'\1a'])
            message = await fivesec(xpub_socket.recv_multipart())
            assert message == [b'\1a']

        finally:
            xpub_socket.close()
            xsub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_rep_invalid_binds(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            # We trigger all kind of binding errors: invalid endpoints,
            # duplicate ports, and so on.
            req_socket.bind('ipc:///non-existing/path')
            req_socket.bind(endpoint)

            # This one is special and fails immediately.
            with pytest.raises(UnsupportedSchemeError):
                req_socket.bind('foo://some-endpoint')

            # This bind is valid and the test should work nonetheless.
            req_socket.bind(endpoint)
            rep_socket.connect(endpoint)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            message = await fivesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']
            await fivesec(rep_socket.send_multipart([b'my', b'response']))
            message = await req_socket.recv_multipart()
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_invalid_operation(event_loop, endpoint):
    async with azmq.Context() as context:
        async with context.socket(azmq.REQ) as req_socket:
            req_socket.connect(endpoint)

            await fivesec(req_socket.send_multipart([b'my', b'request']))

            with pytest.raises(InvalidOperation):
                await fivesec(req_socket.send_multipart([b'my', b'request']))


@use_all_transports
@pytest.mark.asyncio
async def test_rep_invalid_operation(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        rep_socket = context.socket(azmq.REP)

        try:
            req_socket.bind(endpoint)
            rep_socket.connect(endpoint)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            message = await fivesec(rep_socket.recv_multipart())
            assert message == [b'my', b'request']

            with pytest.raises(InvalidOperation):
                await fivesec(rep_socket.recv_multipart())

            await fivesec(rep_socket.send_multipart([b'my', b'response']))
            message = await req_socket.recv_multipart()
            assert message == [b'my', b'response']

        finally:
            req_socket.close()
            rep_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_pub_sub_subscribe_while_connected(event_loop, endpoint):
    async with azmq.Context() as context:
        pub_socket = context.socket(azmq.PUB)
        sub_socket = context.socket(azmq.SUB)
        await sub_socket.subscribe(b'h')

        try:
            pub_socket.bind(endpoint)
            sub_socket.connect(endpoint)

            # Let's start a task that sends messages on the pub socket.
            async def publish():
                while True:
                    await fivesec(pub_socket.send_multipart([b'hello']))
                    await fivesec(pub_socket.send_multipart([b'world']))

            publish_task = asyncio.ensure_future(publish())

            try:
                message = await fivesec(sub_socket.recv_multipart())
                assert message == [b'hello']

                await fivesec(sub_socket.unsubscribe(b'h'))
                await fivesec(sub_socket.subscribe(b'w'))

                while message == [b'hello']:
                    message = await fivesec(sub_socket.recv_multipart())

                assert message == [b'world']
            finally:
                publish_task.cancel()

        finally:
            pub_socket.close()
            sub_socket.close()


@use_all_transports
@pytest.mark.asyncio
async def test_req_router_invalid(event_loop, endpoint):
    async with azmq.Context() as context:
        req_socket = context.socket(azmq.REQ)
        router_socket = context.socket(azmq.ROUTER)

        try:
            req_socket.identity = b'bob'
            req_socket.bind(endpoint)
            router_socket.connect(endpoint)

            await fivesec(req_socket.send_multipart([b'my', b'request']))
            message = await fivesec(router_socket.recv_multipart())
            assert message == [b'bob', b'', b'my', b'request']
            await fivesec(router_socket.send_multipart([b'bob', b'bug']))
            await zerosec(req_socket.recv_multipart())

        finally:
            req_socket.close()
            router_socket.close()



"""
Inproc transport classes.
"""

import asyncio

from ..common import (
    AsyncBox,
    ClosableAsyncObject,
    CompositeClosableAsyncObject,
    cancel_on_closing,
)


class Channel(ClosableAsyncObject):
    def on_open(self, path):
        super().on_open()
        self._path = path
        self._linked_channel = None
        self._inbox = AsyncBox(loop=self.loop)

    async def on_close(self):
        self._inbox.close()

        if self._linked_channel:
            self._linked_channel.close()
            self._linked_channel._linked_channel = None
            self._linked_channel = None

        await self._inbox.wait_closed()

    @property
    def path(self):
        return self._path

    def __repr__(self):
        return 'Channel(path=%r)' % self._path

    def link(self, channel):
        self._linked_channel = channel
        channel._linked_channel = self

    @cancel_on_closing
    async def read(self):
        return await self._inbox.read()

    @cancel_on_closing
    async def write(self, item):
        if self._linked_channel:
            await self._linked_channel._inbox.write(item)


class InprocServer(CompositeClosableAsyncObject):
    def on_open(self, handler):
        super().on_open()
        self._handler = handler
        self._tasks = []

    async def on_close(self):
        if self._tasks:
            await asyncio.wait(self._tasks[:])

        await super().on_close()

    @cancel_on_closing
    async def create_channel(self, path):
        left = Channel(path=path, loop=self.loop)
        self.register_child(left)
        right = Channel(path=path, loop=self.loop)
        self.register_child(right)
        left.link(right)

        task = asyncio.ensure_future(self._handler(right))
        self._tasks.append(task)

        @task.add_done_callback
        def remove_task(future):
            self._tasks.remove(task)

        return left



"""
ZAP-related classes.
"""

import asyncio

from itertools import count

from .constants import (
    DEALER,
    ROUTER,
)
from .common import (
    CompositeClosableAsyncObject,
    cancel_on_closing,
)
from .errors import (
    ZAPTemporaryError,
    ZAPAuthenticationFailure,
    ZAPInternalError,
    ZAPError,
)
from .metadata import (
    buffer_to_metadata,
    metadata_to_buffers,
)

from .log import logger

ZAP_INPROC_ENDPOINT = 'inproc://zeromq.zap.01'


class ZAPClient(CompositeClosableAsyncObject):
    """
    A ZAP client.
    """
    def on_open(self, context):
        super().on_open()

        self._socket = context.socket(DEALER)
        self._socket.connect(ZAP_INPROC_ENDPOINT)
        self.register_child(self._socket)
        self._requests = {}
        self._request_ids = count()
        self._run_task = asyncio.ensure_future(self.run())

    @cancel_on_closing
    async def run(self):
        while True:
            frames = await self._socket.recv_multipart()

            try:
                assert frames[0] == b''
                assert frames[1] == b'1.0'
                request_id = frames[2]
                status_code = int(frames[3].decode('ascii'))
                status_text = frames[4].decode('utf-8')
                user_id = frames[5].decode('utf-8')
                metadata = buffer_to_metadata(frames[6]) if frames[6] else {}
            except Exception as ex:
                logger.warning(
                    "Unexpected error while handling ZAP response (%r): %s."
                    " Ignoring.",
                    frames,
                    ex,
                )
                continue

            future = self._requests.get(request_id)

            if not future:
                logger.debug(
                    "Got authentication result for unknown request %r. "
                    "Ignoring.",
                    request_id,
                )
            else:
                if status_code == 200:
                    future.set_result((user_id, metadata))
                elif status_code == 300:
                    future.set_exception(ZAPTemporaryError(status_text))
                elif status_code == 400:
                    future.set_exception(ZAPAuthenticationFailure(status_text))
                elif status_code == 500:
                    future.set_exception(ZAPInternalError(status_text))
                else:
                    future.set_exception(ZAPError(status_text, status_code))

    def _get_next_request_id(self):
        request_id = next(self._request_ids)
        return '%x' % request_id

    @cancel_on_closing
    async def authenticate(
        self,
        domain,
        address,
        identity,
        mechanism,
        credentials,
    ):
        request_id = self._get_next_request_id()
        await self._socket.send_multipart([
            b'',
            b'1.0',
            request_id,
            domain.encode('utf-8'),
            address.encode('ascii'),
            identity,
            mechanism,
        ] + credentials)
        future = asyncio.Future(loop=self.loop)
        self._requests[request_id] = future

        @future.add_done_callback
        def remove_request(_):
            del self._requests[request_id]

        await future
        return future.result()


class BaseZAPAuthenticator(CompositeClosableAsyncObject):
    """
    A base class for ZAP authenticators.
    """
    def on_open(self, context):
        super().on_open()

        self._socket = context.socket(ROUTER)
        self._socket.bind(ZAP_INPROC_ENDPOINT)
        self.register_child(self._socket)
        self.run_task = asyncio.ensure_future(self.run())

    @cancel_on_closing
    async def run(self):
        logger.debug("ZAP authenticator started.")

        try:
            while True:
                frames = await self._socket.recv_multipart()

                try:
                    delimiter_index = frames.index(b'')
                    envelope = frames[:delimiter_index + 1]
                    message = frames[delimiter_index + 1:]
                    assert message[0] == b'1.0'
                    request_id = message[1]
                    domain = message[2].decode('utf-8')
                    address = message[3].decode('ascii')
                    identity = message[4]
                    mechanism = message[5]
                    credentials = message[6:]

                except Exception as ex:
                    logger.warning(
                        "Unexpected error while handling ZAP request (%r): %s."
                        " Ignoring.",
                        frames,
                        ex,
                    )
                    continue

                try:
                    user_id, metadata = await self.on_request(
                        domain=domain,
                        address=address,
                        identity=identity,
                        mechanism=mechanism,
                        credentials=credentials,
                    )

                    await self._socket.send_multipart(envelope + [
                        b'1.0',
                        request_id,
                        b'200',
                        b'',
                        user_id.encode('utf-8') if user_id else b'',
                        b''.join(metadata_to_buffers(metadata))
                        if metadata else b''
                    ])
                except ZAPError as ex:
                    await self._socket.send_multipart(envelope + [
                        b'1.0',
                        request_id,
                        str(ex.code).encode('ascii'),
                        ex.text.encode('utf-8'),
                        b'',
                        b'',
                    ])
                except Exception as ex:
                    await self._socket.send_multipart(envelope + [
                        b'1.0',
                        request_id,
                        b'500',
                        str(ex).encode('utf-8'),
                        b'',
                        b'',
                    ])
        finally:
            logger.debug("ZAP authenticator stopped.")

    async def on_request(
        self,
        domain,
        address,
        identity,
        mechanism,
        credentials,
    ):
        """
        Handle a ZAP request.
        """
        raise NotImplementedError  # pragma: no cover


class ZAPAuthenticator(BaseZAPAuthenticator):
    """
    A basic ZAP authenticator.
    """
    def on_open(self, *args, **kwargs):
        super().on_open(*args, **kwargs)
        self.whitelist = []
        self.blacklist = []
        self.passwords = {}
        self.authorized_keys = set()

    def allow(self, ip):
        """
        Allow the specified IP to connect.

        :param ip: The IPv4 or IPv6 address to allow.
        """
        self.whitelist.append(ip)
        self.blacklist.clear()

    def deny(self, ip):
        """
        Allow the specified IP to connect.

        :param ip: The IPv4 or IPv6 address to allow.
        """
        self.blacklist.append(ip)
        self.whitelist.clear()

    def add_user(self, username, password):
        """
        Add or update a user in the internal passwords database, for use with
        the PLAIN mechanism.

        :param username: The username.
        :param password: The password.
        """
        self.passwords[username] = password

    def remove_user(self, username):
        """
        Remove a user from the internal passwords database.

        :param username: The username. If no such user exists, nothing is done.
        """
        try:
            del self.passwords[username]
        except KeyError:
            pass

    def add_authorized_key(self, key):
        """
        Add a new authorized permanent client key.

        :param public_key: The public permanent key.
        """
        self.authorized_keys.add(key)

    def remove_authorized_key(self, key):
        """
        Remove an existing authorized permanent client key.

        :param public_key: The public permanent key.
        """
        self.authorized_keys.discard(key)

    async def on_request(
        self,
        domain,
        address,
        identity,
        mechanism,
        credentials,
    ):
        """
        Handle a ZAP request.
        """
        logger.debug(
            "Request in domain %s for %s (%r): %r (%r)",
            domain,
            address,
            identity,
            mechanism,
            credentials,
        )

        user_id = None
        metadata = {}

        if self.whitelist:
            if address not in self.whitelist:
                raise ZAPAuthenticationFailure(
                    "IP address is not in the whitelist",
                )
        elif self.blacklist:
            if address in self.blacklist:
                raise ZAPAuthenticationFailure("IP address is blacklisted")

        if mechanism == b'PLAIN':
            username = credentials[0].decode('utf-8')
            password = credentials[1].decode('utf-8')
            ref_password = self.passwords.get(username)

            if not ref_password:
                raise ZAPAuthenticationFailure("No such user %r" % username)

            if password != ref_password:
                raise ZAPAuthenticationFailure(
                    "Invalid password for user %r" % username,
                )

            user_id = username

        elif mechanism == b'CURVE':
            public_key = credentials[0]

            if public_key not in self.authorized_keys:
                raise ZAPAuthenticationFailure(
                    "Unauthorized key %r" % public_key,
                )

        return user_id, metadata



"""
The PLAIN client mechanism.
"""

import struct

from ..log import logger

from .base import Mechanism


class PlainClient(object):
    def __init__(self, username, password):
        assert len(username) < 256, "username can't exceed 255 bytes"
        assert len(password) < 256, "password can't exceed 255 bytes"

        self.username = username
        self.password = password

    def __call__(self):
        return PlainClientMechanism(
            username=self.username,
            password=self.password,
        )


class PlainClientMechanism(Mechanism):
    name = b'PLAIN'
    as_server = False

    __slots__ = [
        'username',
        'password',
    ]

    def __init__(self, username, password):
        super().__init__()
        self.username = username
        self.password = password

    def _write_plain_hello(self, writer):
        self.write_command(
            writer=writer,
            name=b'HELLO',
            buffers=[
                struct.pack('B', len(self.username)),
                self.username.encode('utf-8'),
                struct.pack('B', len(self.password)),
                self.password.encode('utf-8'),
            ],
        )

    @classmethod
    def _write_plain_initiate(cls, writer, metadata):
        cls.write_command(
            writer=writer,
            name=b'INITIATE',
            buffers=cls._metadata_to_buffers(metadata)
        )

    @classmethod
    async def _read_plain_ready(cls, reader):
        raw_metadata = await cls._expect_command(
            reader=reader,
            name=b'READY',
        )
        return cls._buffer_to_metadata(buffer=raw_metadata)

    async def negotiate(self, writer, reader, metadata, address, zap_client):
        logger.debug("Negotiating PLAIN parameters as client.")

        self._write_plain_hello(writer=writer)
        await self._expect_command(reader=reader, name=b'WELCOME')
        self._write_plain_initiate(writer=writer, metadata=metadata)
        return await self._read_plain_ready(reader=reader), None, None
